/*
 *           Grain - a minimal windowing system, version 1.1
 *
 *                            Written by
 *                          Harald Winroth
 *      Computational Vision and Active Perception Lab (CVAP)
 *                  Royal Institute of Technology
 *                            Stockholm
 * 
 *                Copyright (c) 1998 Harald Winroth
 *
 * See the file "Copying" for information on usage and redistribution of 
 * this file, and for a disclaimer of all warranties. Please send bug 
 * reports and direct all inquiries to harald@nada.kth.se.
 *
 */

#ifndef GRAIN_WINDOW_HH
#define GRAIN_WINDOW_HH

#include <grain/geometry.hh>
#include <cstddef>
#include <vector>
#include <set>
#include <string> 

struct GrainColor {
    GrainColor(); // creates a transparent color
    GrainColor(unsigned char red, unsigned char green, unsigned char blue);
    bool operator==(const GrainColor& color) const;
    bool operator<(const GrainColor& color) const;
    unsigned char r, g, b;
    bool fill;

    // Pre-defined colors
    static const GrainColor transparent;
    static const GrainColor black;
    static const GrainColor white;
    static const GrainColor red;
    static const GrainColor green;
    static const GrainColor blue;
};

class GrainWindowEventHandler;
class GrainWindowingSystem;

class GrainWindow
{
public:
    GrainWindow();
    virtual ~GrainWindow() {}

    void set_event_handler(GrainWindowEventHandler* event_handler);
    GrainWindowEventHandler* event_handler() const;

    virtual GrainWindowingSystem& system() const=0;

    virtual void set_bounds(const GrainRect& bounds)=0;
    virtual GrainRect bounds() const=0;
    
    virtual GrainCoord absolute(const GrainCoord& pos) const=0;
    virtual GrainCoord relative(const GrainCoord& pos) const=0;

    virtual void refresh(const GrainRect& area)=0;
    void refresh();
    virtual void draw_point(const GrainCoord& at, 
			    const GrainColor& color=GrainColor::black)=0;
    virtual void draw_points(const vector<GrainCoord>& at, 
			     const GrainColor& color=GrainColor::black)=0;
    virtual void draw_line(const GrainCoord& from, const GrainCoord& to, 
			   const GrainColor& color=GrainColor::black, 
			   size_t line_width=1)=0;
    virtual void draw_rectangle(const GrainRect& bounds, 
				const GrainColor& border_color=
				    GrainColor::black,
				const GrainColor& fill_color=
				    GrainColor::transparent, 
				size_t line_width=1)=0;
    virtual void draw_text(const GrainCoord& at, const string& text, 
			   const string& font_name, 
			   const GrainColor& color=GrainColor::black)=0;

    virtual GrainWindow* create_subwindow(const GrainRect& bounds)=0;

    virtual void grab_pointer() const=0;
    virtual void ungrab_pointer() const=0;
    virtual void grab_keyboard() const=0;
    virtual void ungrab_keyboard() const=0;

private:
    GrainWindowEventHandler* event_handler_;
};

class GrainWindowEventHandler
{
public:
    virtual ~GrainWindowEventHandler() {}

    enum Modifier {
	Shift, 
	Control,
	Meta,
	Alt
    };

    virtual void press_event(GrainWindow* window, const GrainCoord& pos, 
			     size_t button, const set<Modifier>& modifiers) {}
    virtual void release_event(GrainWindow* window, const GrainCoord& pos, 
			       size_t button) {}
    virtual void motion_event(GrainWindow* window, const GrainCoord& pos) {}
    virtual void enter_event(GrainWindow* window, const GrainCoord& pos) {}
    virtual void leave_event(GrainWindow* window, const GrainCoord& pos) {}
    virtual void key_event(GrainWindow* window, const string& value, 
			   const set<Modifier>& modifiers) {}
    virtual void refresh_event(GrainWindow* window, const GrainRect& bounds) {}
};

class GrainWindowingSystem
{
public:
    class Exception : public exception 
    {
    public:
	Exception(const string& what);

	// Implementation of interface exception
    public:
	const char* what () const;

    private:
	string what_;
    };

    virtual ~GrainWindowingSystem() {}

    virtual GrainWindow* create_base_window(const GrainRect& bounds, 
					    bool resizable, 
					    const string& title)=0;
    virtual GrainWindow* create_pop_up_window(const GrainRect& bounds)=0;

    virtual GrainCoord text_size(const string& font_name, 
				 const string& text) const=0;
    virtual size_t text_index(size_t offset, const string& font_name, 
			      const string& text) const=0;

    virtual void enter_event_loop()=0;
    virtual void exit_event_loop()=0;
    static void error(const string& what);
};

#endif // GRAIN_WINDOW_HH
