/*
 *           Grain - a minimal windowing system, version 1.1
 *
 *                            Written by
 *                          Harald Winroth
 *      Computational Vision and Active Perception Lab (CVAP)
 *                  Royal Institute of Technology
 *                            Stockholm
 * 
 *                Copyright (c) 1998 Harald Winroth
 *
 * See the file "Copying" for information on usage and redistribution of 
 * this file, and for a disclaimer of all warranties. Please send bug 
 * reports and direct all inquiries to harald@nada.kth.se.
 *
 */

#ifndef GRAIN_X_WINDOW_HH
#define GRAIN_X_WINDOW_HH

#include <grain/window.hh>
#include <vector>
#include <map>
#include <set>
#include <iterator>
extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
}

class GrainXWindowingSystem : public GrainWindowingSystem
{
    friend class GrainXWindow;
    friend class GrainXBaseWindow;
    friend class GrainXPopUpWindow;
    friend class GrainXSubwindow;

public:
    GrainXWindowingSystem();
    GrainXWindowingSystem(const string& display_name);
    ~GrainXWindowingSystem();

private:
    // Interface for GrainXWindow and its subclasses

    Display* display() const;
    int screen() const;
    Colormap colormap() const;
    GC gc() const;
    unsigned long pixel(const GrainColor& color);
    void draw_text(Window window, const GrainCoord& at, const string& text, 
		   const string& font_name, const GrainColor& color);
    void add_window(GrainXWindow* window, Window id);
    void remove_window(GrainXWindow* window, Window id);
    GrainXWindow* window(const XEvent& event) const;

private: // the *really* private section
    enum LockUsage {
	LockUsageIgnore,
	LockUsageShift,
	LockUsageCaps
    };

    GrainXWindowingSystem(const GrainXWindowingSystem&); // undefined
    GrainXWindowingSystem& operator=(
	const GrainXWindowingSystem&); // undefined
    void open_display(const char* display_name);
    void set_font(const string& font_name) const;
    void init_keymap_info();
    string key(XKeyEvent* key_event);
    set<GrainWindowEventHandler::Modifier> modifiers(unsigned int state);
    static unsigned short scale_color(unsigned char c);
    
    Display* display_;
    int screen_;
    Colormap colormap_;
    GC gc_;
    mutable string font_name_;
    mutable XFontStruct* font_; // if !=0, it contains info about font_name_
    bool exit_loop_;
    vector<GrainColor> colors_;
    vector<unsigned long> pixels_;
    map<Window, GrainXWindow*> windows_;
    LockUsage lock_usage_;
    unsigned int mode_mod_mask_, meta_mod_mask_, alt_mod_mask_;

    // Implementation of interface GrainWindowingSystem
public:    
    GrainWindow* create_base_window(const GrainRect& bounds, bool resizable, 
				    const string& title);
    GrainWindow* create_pop_up_window(const GrainRect& bounds);
    GrainCoord text_size(const string& font_name, const string& text) const;
    size_t text_index(size_t offset, const string& font_name, 
		      const string& text) const;
    void enter_event_loop();
    void exit_event_loop();
};

class GrainXWindow : public GrainWindow
{
public:
    ~GrainXWindow();

protected:
    GrainXWindow(GrainXWindowingSystem& system, Window window_id);
    Window window_id() const;

private:
    GrainXWindow(const GrainXWindow&); // undefined
    GrainXWindow& operator=(const GrainXWindow&); // undefined

    GrainXWindowingSystem& system_;
    Window window_id_;

    // Implementation of interface GrainWindow
public:
    GrainXWindowingSystem& system() const;

    void set_bounds(const GrainRect& bounds);
    GrainRect bounds() const;

    GrainCoord absolute(const GrainCoord& pos) const;
    GrainCoord relative(const GrainCoord& pos) const;

    void refresh(const GrainRect& area);
    void draw_point(const GrainCoord& at, const GrainColor& color);
    void draw_points(const vector<GrainCoord>& at, const GrainColor& color);
    void draw_line(const GrainCoord& from, const GrainCoord& to, 
		   const GrainColor& color, size_t line_width);
    void draw_rectangle(const GrainRect& bounds, 
			const GrainColor& border_color,
			const GrainColor& fill_color, size_t line_width);
    void draw_text(const GrainCoord& at, const string& text, 
		   const string& font_name, const GrainColor& color);
    GrainWindow* create_subwindow(const GrainRect& bounds);
    void grab_pointer() const;
    void ungrab_pointer() const;
    void grab_keyboard() const;
    void ungrab_keyboard() const;
};

class GrainXBaseWindow : public GrainXWindow
{
    friend GrainWindow* GrainXWindowingSystem::create_base_window(
	const GrainRect& bounds, bool resizable, const string& title);

protected:
    GrainXBaseWindow(GrainXWindowingSystem& system, const GrainRect& bounds, 
		bool resizable, const string& title);

private:
    Window create_x_window(Display* display, int screen, 
			   const GrainRect& bounds, bool resizable, 
			   const string& title);
};

class GrainXPopUpWindow : public GrainXWindow
{
    friend GrainWindow* GrainXWindowingSystem::create_pop_up_window(
	const GrainRect& bounds);

protected:
    GrainXPopUpWindow(GrainXWindowingSystem& system, const GrainRect& bounds);

private:
    Window create_x_window(Display* display, int screen, 
			   const GrainRect& bounds);
};

class GrainXSubwindow : public GrainXWindow
{
    friend GrainWindow* GrainXWindow::create_subwindow(
	const GrainRect& bounds);

protected:
    GrainXSubwindow(GrainXWindowingSystem& system, Window parent, 
		    const GrainRect& bounds);

private:
    Window create_x_window(Display* display, int screen, Window parent, 
			   const GrainRect& bounds);
};

#endif // GRAIN_X_WINDOW_HH
