/*
 *           Grain - a minimal windowing system, version 1.1
 *
 *                            Written by
 *                          Harald Winroth
 *      Computational Vision and Active Perception Lab (CVAP)
 *                  Royal Institute of Technology
 *                            Stockholm
 * 
 *                Copyright (c) 1998 Harald Winroth
 *
 * See the file "Copying" for information on usage and redistribution of 
 * this file, and for a disclaimer of all warranties. Please send bug 
 * reports and direct all inquiries to harald@nada.kth.se.
 *
 */

#include <grain/geometry.hh>
#include <algorithm>

GrainCoord::GrainCoord() :
    x(0), y(0)
{}

GrainCoord::GrainCoord(long xx, long yy) :
    x(xx), y(yy)
{}

bool GrainCoord::operator==(const GrainCoord& coord) const
{
    return x == coord.x && y == coord.y;
}

GrainCoord operator+(const GrainCoord& c1, const GrainCoord& c2)
{
    return GrainCoord(c1.x+c2.x, c1.y+c2.y);
}

GrainCoord operator-(const GrainCoord& c1, const GrainCoord& c2)
{
    return GrainCoord(c1.x-c2.x, c1.y-c2.y);
}

GrainCoord& operator+=(GrainCoord& c1, const GrainCoord& c2)
{
    c1.x+=c2.x;
    c1.y+=c2.y;
    return c1;
}

GrainCoord& operator-=(GrainCoord& c1, const GrainCoord& c2)
{
    c1.x-=c2.x;
    c1.y-=c2.y;
    return c1;
}

ostream& operator<<(ostream& os, const GrainCoord& c)
{
    os << c.x << " " << c.y;
    return os;
}

GrainRect::GrainRect()
{}

GrainRect::GrainRect(const GrainCoord& pos, const GrainCoord& sz) :
    position(pos), size(sz)
{
    // Precondition: sz.x >=0 && sz.y >=0
}

GrainRect GrainRect::with_corners(const GrainCoord& corner1, 
				  const GrainCoord& corner2)
{
    GrainCoord c1(corner1);
    GrainCoord c2(corner2);
    if (c1.x > c2.x)
	swap(c1.x, c2.x);
    if (c1.y > c2.y)
	swap(c1.y, c2.y);
    return GrainRect(c1, c2-c1);
}

bool GrainRect::operator==(const GrainRect& rect) const
{
    return position == rect.position && size == rect.size;
}

GrainRect intersection(const GrainRect& r1, const GrainRect& r2)
{
    GrainCoord c11=r1.position;
    GrainCoord c12=c11+r1.size;
    GrainCoord c21=r2.position;
    GrainCoord c22=c21+r2.size;

    if (c11.x<c21.x)
	c11.x=c21.x;
    if (c11.y<c21.y)
	c11.y=c21.y;
    if (c12.x>c22.x)
	c12.x=c22.x;
    if (c12.y>c22.y)
	c12.y=c22.y;

    GrainRect result(c11, c12-c11);
    if (result.size.x<0 || result.size.y<0)
	result.size=GrainCoord(0, 0);
    return result;
}

bool intersects(const GrainRect& r1, const GrainRect& r2)
{
    return intersection(r1, r2).size != GrainCoord(0, 0);
}

bool inside(const GrainCoord& p, const GrainRect& r)
{
    GrainCoord c1=r.position;
    GrainCoord c2=c1+r.size;
    return p.x >= c1.x && p.x <= c2.x && p.y >= c1.y && p.y <= c2.y;
}

GrainRect hull(const GrainRect& r1, const GrainRect& r2)
{
    GrainCoord c11=r1.position;
    GrainCoord c12=c11+r1.size;
    GrainCoord c21=r2.position;
    GrainCoord c22=c21+r2.size;

    if (c11.x>c21.x)
	c11.x=c21.x;
    if (c11.y>c21.y)
	c11.y=c21.y;
    if (c12.x<c22.x)
	c12.x=c22.x;
    if (c12.y<c22.y)
	c12.y=c22.y;
    
    return GrainRect::with_corners(c11, c12);
}

GrainRect expand(const GrainRect& r, int steps)
{
    GrainCoord c1=r.position-GrainCoord(steps, steps);
    GrainCoord c2=c1+r.size+GrainCoord(steps, steps);
    return GrainRect::with_corners(c1, c2);
}

ostream& operator<<(ostream& os, const GrainRect& r)
{
    os << r.position << " " << r.size;
    return os;
}
