/*
 *           Grain - a minimal windowing system, version 1.1
 *
 *                            Written by
 *                          Harald Winroth
 *      Computational Vision and Active Perception Lab (CVAP)
 *                  Royal Institute of Technology
 *                            Stockholm
 * 
 *                Copyright (c) 1998 Harald Winroth
 *
 * See the file "Copying" for information on usage and redistribution of 
 * this file, and for a disclaimer of all warranties. Please send bug 
 * reports and direct all inquiries to harald@nada.kth.se.
 *
 */

#include <grain/x-window.hh>
#include <algorithm>
#include <assert.h>
extern "C" {
#include <X11/keysym.h>
#include <X11/cursorfont.h>
}

GrainXWindowingSystem::GrainXWindowingSystem()
{
    open_display(0);
}

GrainXWindowingSystem::GrainXWindowingSystem(const string& display_name)
{
    open_display(display_name.c_str());
}

GrainXWindowingSystem::~GrainXWindowingSystem()
{
    set_font("");
    XFreeGC(display(), gc());
    unsigned long* pixel_array=new unsigned long[pixels_.size()];
    copy(pixels_.begin(), pixels_.end(), pixel_array);
    XFreeColors(display(), colormap(), pixel_array, pixels_.size(), 0);
    delete [] pixel_array;
    XCloseDisplay(display());
}

Display* GrainXWindowingSystem::display() const
{
    return display_;
}

int GrainXWindowingSystem::screen() const
{
    return screen_;
}

Colormap GrainXWindowingSystem::colormap() const
{
    return colormap_;
}

GC GrainXWindowingSystem::gc() const
{
    return gc_;
}

unsigned long GrainXWindowingSystem::pixel(const GrainColor& color)
{
    vector<GrainColor>::iterator i=find(colors_.begin(), colors_.end(), color);
    if (i != colors_.end())
	return pixels_[i-colors_.begin()];

    XColor xcolor;
    xcolor.red=scale_color(color.r);
    xcolor.green=scale_color(color.g);
    xcolor.blue=scale_color(color.b);
    xcolor.flags=DoRed|DoGreen|DoBlue;
    if (!XAllocColor(display(), colormap(), &xcolor))
	error("GrainXWindowingSystem::pixel: could not allocate "
	      "requested color");
    colors_.push_back(color);
    pixels_.push_back(xcolor.pixel);
    return xcolor.pixel;
}

void GrainXWindowingSystem::draw_text(
    Window window, const GrainCoord& at, const string& text, 
    const string& font_name, const GrainColor& color)
{
    set_font(font_name);
    XSetForeground(display(), gc(), pixel(color));
    XDrawString(display(), window, gc(), at.x, at.y+font_->ascent, 
		const_cast<char*>(text.c_str()), text.length());
}

void GrainXWindowingSystem::add_window(GrainXWindow* window, Window id)
{
    windows_.insert(make_pair(id, window));
}

void GrainXWindowingSystem::remove_window(GrainXWindow* window, Window id)
{
    map<Window, GrainXWindow*>::iterator i=windows_.find(id);
    assert(i != windows_.end() && (*i).second == window);
    windows_.erase(i);
}

GrainXWindow* GrainXWindowingSystem::window(const XEvent& event) const
{
    map<Window, GrainXWindow*>::const_iterator i=
	windows_.find(event.xany.window);
    if (i != windows_.end())
	return (*i).second;
    return 0;
}

void GrainXWindowingSystem::open_display(const char* display_name)
{
    display_=XOpenDisplay(const_cast<char*>(display_name));
    if (display_ == 0)
	error(string("GrainXWindowingSystem: cannot connect to X server ")+
	      XDisplayName(display_name));
    screen_=DefaultScreen(display());
    colormap_=DefaultColormap(display(), screen());
    gc_=XCreateGC(display(), RootWindow(display(), screen()), 0, 0);
    XSetFunction(display(), gc_, GXcopy);
    font_=0;
    exit_loop_=false;
    init_keymap_info();
}

void GrainXWindowingSystem::set_font(const string& font_name) const
{
    if (font_name==font_name_)
	return;
    if (font_)
	XFreeFont(display(), font_);
    font_name_="";
    if (font_name=="")
	return;
    font_=XLoadQueryFont(display(), const_cast<char*>(font_name.c_str()));
    if (font_==0)
	error("GrainXWindowingSystem::set_font: failed to load font "+
	      font_name);
    font_name_=font_name;
    XSetFont(display(), gc(), font_->fid);
}

void GrainXWindowingSystem::init_keymap_info()
{
    XModifierKeymap* mod_map=XGetModifierMapping(display());

    // Check the keycodes associated with the Lock modifier. If
    // any of them is associated with the XK_Shift_Lock modifier,
    // then Lock has to be interpreted as Shift Lock, not Caps Lock.
    // LockMapIndex is defined by X in X11/X.h.
    lock_usage_=LockUsageIgnore;
    KeyCode* code=mod_map->modifiermap+LockMapIndex*mod_map->max_keypermod;
    for (int count=mod_map->max_keypermod; count>0; --count, ++code) {
	if (*code==0)
	    continue;
	KeySym keysym=XKeycodeToKeysym(display(), *code, 0);
	if (keysym==XK_Shift_Lock) {
	    lock_usage_=LockUsageShift;
	    break;
	}
	if (keysym==XK_Caps_Lock) {
	    lock_usage_=LockUsageCaps;
	    break;
	}
    }

    // Look through the keycodes associated with modifiers to see if
    // the the "mode switch", "meta", or "alt" keysyms are associated
    // with any modifiers. If so, remember their modifier mask bits.
    mode_mod_mask_=0;
    meta_mod_mask_=0;
    alt_mod_mask_=0;
    for (int map_idx=0; map_idx<8; ++map_idx) {
	// map_idx is one of ShiftMapIndex, LockMapIndex etc. See
	// X11/X.h for details.
	code=mod_map->modifiermap+map_idx*mod_map->max_keypermod;
	for (int i=0; i<mod_map->max_keypermod; ++i, ++code) {
	    unsigned int mask=(ShiftMask << map_idx);
	    // mask is one of ShiftMask, LockMask etc. See X11/X.h for details.
	    if (*code==0)
		continue;
	    KeySym keysym=XKeycodeToKeysym(display(), *code, 0);
	    if (keysym==XK_Mode_switch)
		mode_mod_mask_ |= mask;
	    if ((keysym == XK_Meta_L) || (keysym == XK_Meta_R))
		meta_mod_mask_ |= mask;
	    if ((keysym == XK_Alt_L) || (keysym == XK_Alt_R))
		alt_mod_mask_ |= mask;
	}
    }

    XFreeModifiermap(mod_map);
}

string GrainXWindowingSystem::key(XKeyEvent* key_event)
{
    // Figure out which of the four slots in the keymap vector to
    // use for this key.  Refer to Xlib documentation for more info
    // on how this computation works.

    int idx=0;
    if (key_event->state & mode_mod_mask_)
	idx=2;
    if ((key_event->state & ShiftMask)
	|| ((lock_usage_ != LockUsageIgnore) 
	    && (key_event->state & LockMask)))
    {
	idx+=1;
    }

    KeySym sym=XKeycodeToKeysym(display(), key_event->keycode, idx);

    // Special handling:  if the key was shifted because of Lock, but
    // lock is only caps lock, not shift lock, and the shifted keysym
    // isn't upper-case alphabetic, then switch back to the unshifted
    // keysym.
    if ((idx & 1) && !(key_event->state & ShiftMask)
	&& (lock_usage_==LockUsageCaps)) 
    {
	if (!(((sym >= XK_A) && (sym <= XK_Z))
	      || ((sym >= XK_Agrave) && (sym <= XK_Odiaeresis))
	      || ((sym >= XK_Ooblique) && (sym <= XK_Thorn)))) 
	{
	    idx &= ~1;
	    sym=XKeycodeToKeysym(display(), key_event->keycode, idx);
	}
    }

    // Another bit of special handling:  if this is a shifted key and there
    // is no keysym defined, then use the keysym for the unshifted key.

    if ((idx & 1) && (sym == NoSymbol)) {
	sym=XKeycodeToKeysym(display(), key_event->keycode, idx & ~1);
    }

    return XKeysymToString(sym);
}

set<GrainWindowEventHandler::Modifier> 
GrainXWindowingSystem::modifiers(unsigned int state)
{
    set<GrainWindowEventHandler::Modifier> mod;
    if (state & ShiftMask)
	mod.insert(GrainWindowEventHandler::Shift);
    if (state & ControlMask)
	mod.insert(GrainWindowEventHandler::Control);
    if (state & meta_mod_mask_)
	mod.insert(GrainWindowEventHandler::Meta);
    if (state & alt_mod_mask_)
	mod.insert(GrainWindowEventHandler::Alt);
    return mod;
}

unsigned short GrainXWindowingSystem::scale_color(unsigned char c)
{
    return (unsigned short)(((double)c)*65535.0/255.0+0.5);
}

GrainWindow* GrainXWindowingSystem::create_base_window(
    const GrainRect& bounds, bool resizable, const string& title)
{
    return new GrainXBaseWindow(*this, bounds, resizable, title);
}

GrainWindow* GrainXWindowingSystem::create_pop_up_window(
    const GrainRect& bounds)
{
    return new GrainXPopUpWindow(*this, bounds);
}

GrainCoord GrainXWindowingSystem::text_size(const string& font_name, 
					    const string& text) const
{
    set_font(font_name);
    int direction, ascent, descent;
    XCharStruct overall;
    XTextExtents(font_, text.c_str(), text.length(), 
		 &direction, &ascent, &descent, &overall);
    return GrainCoord(overall.width, ascent+descent);
}
    
size_t GrainXWindowingSystem::text_index(
    size_t offset, const string& font_name, const string& text) const
{
    // This is a lazy implementation; it could be done more efficiently.
    for (size_t i=0; i<=text.length(); ++i)
	if (text_size(font_name, text.substr(0, i)).x > (long)offset)
	    return i;
    return text.length()+1;
}

void GrainXWindowingSystem::enter_event_loop()
{
    exit_loop_=false;
    while (!exit_loop_) {
	XEvent event;
	XNextEvent(display(), &event);
	GrainXWindow* w=window(event);
	if (!w) {
	    // No target window found. It has probably been deleted since
	    // the event entered the queue; just ignore this event.
	    continue;
	}
	GrainWindowEventHandler* eh=w->event_handler();
	if (!eh) {
	    // No event hander installed for the target window; ignore the 
	    // event.
	    continue;
	}
	switch (event.type) {
	case ButtonPress:
	    eh->press_event(w, GrainCoord(event.xbutton.x, event.xbutton.y), 
			    event.xbutton.button, 
			    modifiers(event.xbutton.state));
	    break;
	case ButtonRelease:
	    eh->release_event(w, GrainCoord(event.xbutton.x, event.xbutton.y), 
			      event.xbutton.button);
	    break;
	case MotionNotify:
	    eh->motion_event(w, GrainCoord(event.xmotion.x, event.xmotion.y));
	    break;
	case EnterNotify:
	    eh->enter_event(w, GrainCoord(event.xcrossing.x, 
					  event.xcrossing.y));
	    break;
	case LeaveNotify:
	    eh->leave_event(w, GrainCoord(event.xcrossing.x, 
					  event.xcrossing.y));
	    break;
	case KeyPress:
	    eh->key_event(w, key(&event.xkey), modifiers(event.xkey.state));
	    break;
	case ConfigureNotify:
	    // Just generate an exposure event for the entire window
	    XClearArea(display(), event.xany.window, 0, 0, 
		       event.xconfigure.width, event.xconfigure.height, 
		       /* exposures */ 1);
	    break;
	case Expose:
	    eh->refresh_event(
		w, GrainRect(GrainCoord(event.xexpose.x, event.xexpose.y), 
			     GrainCoord(event.xexpose.width, 
					event.xexpose.height)));
	    break;
	}
    }
}

void GrainXWindowingSystem::exit_event_loop()
{
    exit_loop_=true;
}

GrainXWindow::~GrainXWindow()
{
    system().remove_window(this, window_id_);
    XDestroyWindow(system().display(), window_id_);
}

GrainXWindow::GrainXWindow(GrainXWindowingSystem& system, Window window_id) :
    system_(system),
    window_id_(window_id)
{
    XSelectInput(system.display(), window_id_, 
		 ButtonPressMask |
		 ButtonReleaseMask |
		 EnterWindowMask |
		 LeaveWindowMask |
		 PointerMotionMask |
		 KeyPressMask |
		 ExposureMask |
		 StructureNotifyMask);
    XMapWindow(system.display(), window_id_);
    system.add_window(this, window_id);
}

Window GrainXWindow::window_id() const
{
    return window_id_;
}

GrainXWindowingSystem& GrainXWindow::system() const
{
    return system_;
}

void GrainXWindow::set_bounds(const GrainRect& bounds)
{
    unsigned int value_mask=CWX|CWY|CWWidth|CWHeight;
    XWindowChanges values;
    values.x=bounds.position.x;
    values.y=bounds.position.y;
    values.width=bounds.size.x;
    values.height=bounds.size.y;
    XConfigureWindow(system().display(), window_id_, value_mask, &values);
}

GrainRect GrainXWindow::bounds() const
{
    Window root;
    int x, y;
    unsigned int width, height, border_width, depth;
    if (!XGetGeometry(system().display(), window_id_, &root, &x, &y, &width,
		      &height, &border_width, &depth)) {
	system().error("GrainXWindow::bounds: failed to get window geometry");
    }
    return GrainRect(GrainCoord(x, y), GrainCoord(width, height));
}

GrainCoord GrainXWindow::absolute(const GrainCoord& pos) const
{
    Window root=RootWindow(system().display(), system().screen());
    Window child;
    int x, y;
    if (!XTranslateCoordinates(system().display(), window_id_, root,
			       pos.x, pos.y, &x, &y, &child)) 
    {
	system().error("GrainXWindow::absolute: failed to translate "
		       "coordinates");
    }
    return GrainCoord(x, y);
}

GrainCoord GrainXWindow::relative(const GrainCoord& pos) const
{
    Window root=RootWindow(system().display(), system().screen());
    Window child;
    int x, y;
    if (!XTranslateCoordinates(system().display(), root, window_id_, 
			       pos.x, pos.y, &x, &y, &child)) 
    {
	system().error("GrainXWindow::relative: failed to translate "
		       "coordinates");
    }
    return GrainCoord(x, y);
}

void GrainXWindow::refresh(const GrainRect& area)
{
    if (area.size.x == 0 || area.size.y == 0)
	return;
    XClearArea(system().display(), window_id_, area.position.x, 
	       area.position.y, area.size.x, area.size.y, True);
}

void GrainXWindow::draw_point(const GrainCoord& at, const GrainColor& color)
{
    if (color != GrainColor::transparent) {
	XSetForeground(system().display(), system().gc(), 
		       system().pixel(color));
	XDrawPoint(system().display(), window_id_, system().gc(), at.x, at.y);
    }
}

void GrainXWindow::draw_points(const vector<GrainCoord>& at, 
			       const GrainColor& color)
{
    if (color != GrainColor::transparent) {
	XSetForeground(system().display(), system().gc(), 
		       system().pixel(color));
	XPoint* array=new XPoint[at.size()];
	XPoint* p=array;
	for (vector<GrainCoord>::const_iterator i=at.begin(); i!=at.end(); 
	     ++i, ++p)
	{
	    p->x=(*i).x;
	    p->y=(*i).y;
	}
	XDrawPoints(system().display(), window_id_, system().gc(), array, 
		    at.size(), CoordModeOrigin);
	delete [] array;
    }
}

void GrainXWindow::draw_line(const GrainCoord& from, const GrainCoord& to, 
			const GrainColor& color, size_t line_width)
{
    if (color != GrainColor::transparent) {
	XSetForeground(system().display(), system().gc(), 
		       system().pixel(color));
	XSetLineAttributes(system().display(), system().gc(), 
			   /* unsigned int line_width */ line_width,
			   /* int line_style */ LineSolid,
			   /* int cap_style */ CapButt,
			   /* int join_style */ JoinMiter);
	XDrawLine(system().display(), window_id_, system().gc(), 
		  from.x, from.y, to.x, to.y);
    }
}

void GrainXWindow::draw_rectangle(
    const GrainRect& bounds, const GrainColor& border_color,
    const GrainColor& fill_color, size_t line_width)
{
    if (bounds.size.x > 0 && bounds.size.y > 0 && 
	border_color != GrainColor::transparent)
    {
	XSetForeground(system().display(), system().gc(), 
		       system().pixel(border_color));
	XSetLineAttributes(system().display(), system().gc(), 
			   /* unsigned int line_width */ line_width,
			   /* int line_style */ LineSolid,
			   /* int cap_style */ CapButt,
			   /* int join_style */ JoinMiter);
	XDrawRectangle(system().display(), window_id_, system().gc(), 
		       bounds.position.x, bounds.position.y, 
		       bounds.size.x-1, bounds.size.y-1);
	
    }
    if (bounds.size.x > 2 && bounds.size.y > 2 && 
	fill_color != GrainColor::transparent)
    {
	XSetForeground(system().display(), system().gc(), 
		       system().pixel(fill_color));
	XFillRectangle(system().display(), window_id_, system().gc(), 
		       bounds.position.x+1, bounds.position.y+1, 
		       bounds.size.x-2, bounds.size.y-2);
    }
}

void GrainXWindow::draw_text(const GrainCoord& at, const string& text, 
			const string& font_name, const GrainColor& color)
{
    system().draw_text(window_id_, at, text, font_name, color);
}

GrainWindow* GrainXWindow::create_subwindow(const GrainRect& bounds)
{
    return new GrainXSubwindow(system(), window_id_, bounds);
}

void GrainXWindow::grab_pointer() const
{
    XGrabPointer(
	system().display(),
	/* grab window */ window_id_,
	/* owner events */ False,
	ButtonPressMask |
	ButtonReleaseMask |
	EnterWindowMask |
	LeaveWindowMask |
	PointerMotionMask,
	/* pointer mode */ GrabModeAsync,
	/* keyboard mode */ GrabModeAsync,
	/* confine to */ None,
	/* cursor */ None,
	/* time */ CurrentTime);
}

void GrainXWindow::ungrab_pointer() const
{
    XUngrabPointer(system().display(), CurrentTime);
}

void GrainXWindow::grab_keyboard() const
{
    XGrabKeyboard(
	system().display(),
	/* grab window */ window_id_,
	/* owner events */ False,
	/* pointer mode */ GrabModeAsync,
	/* keyboard mode */ GrabModeAsync,
	/* time */ CurrentTime);
}

void GrainXWindow::ungrab_keyboard() const
{
    XUngrabKeyboard(system().display(), CurrentTime);
}

GrainXBaseWindow::GrainXBaseWindow(
    GrainXWindowingSystem& system, const GrainRect& bounds, bool resizable, 
    const string& title) :
    GrainXWindow(system, create_x_window(
        system.display(), system.screen(), bounds, resizable, title))
{}

Window GrainXBaseWindow::create_x_window(
    Display* display, int screen, const GrainRect& bounds, bool resizable, 
    const string& title)
{
    XSetWindowAttributes attributes;
    attributes.background_pixel=WhitePixel(display, screen);
    attributes.border_pixel=BlackPixel(display, screen);

    Window window_id=XCreateWindow(
        display, 
	/* parent window */ RootWindow(display, screen), 
	bounds.position.x, bounds.position.y, bounds.size.x, bounds.size.y,
        /* border width */ 0, 
	/* depth */ CopyFromParent,
	/* class */ CopyFromParent, 
	/* visual */ CopyFromParent,
	/* value mask */ CWBackPixel|CWBorderPixel,
	&attributes);

    char* window_name=const_cast<char*>(title.c_str());
    char* argv[1]={0};
    XSizeHints size_hints;
    size_hints.flags=PPosition|PSize;
    size_hints.x=bounds.position.x;
    size_hints.y=bounds.position.y;
    size_hints.width=bounds.size.x;
    size_hints.height=bounds.size.y;
    XSetStandardProperties(display, window_id, window_name, window_name, None,
			   argv, 0, &size_hints);

    return window_id;
}

GrainXPopUpWindow::GrainXPopUpWindow(
    GrainXWindowingSystem& system, const GrainRect& bounds) :
    GrainXWindow(system, create_x_window(system.display(), system.screen(), 
 					 bounds))
{}

Window GrainXPopUpWindow::create_x_window(
    Display* display, int screen, const GrainRect& bounds)
{
    XSetWindowAttributes attributes;
    attributes.background_pixel=WhitePixel(display, screen);
    attributes.border_pixel=BlackPixel(display, screen);
    attributes.override_redirect=True;
    attributes.cursor=XCreateFontCursor(display, XC_arrow);

    Window window_id=XCreateWindow(
        display, 
	/* parent window */ RootWindow(display, screen), 
	bounds.position.x, bounds.position.y, bounds.size.x, bounds.size.y,
        /* border width */ 0, 
	/* depth */ CopyFromParent,
	/* class */ CopyFromParent, 
	/* visual */ CopyFromParent,
	/* value mask */ CWBackPixel|CWBorderPixel|CWOverrideRedirect|CWCursor,
	&attributes);

    return window_id;
}


GrainXSubwindow::GrainXSubwindow(GrainXWindowingSystem& system, Window parent, 
		       const GrainRect& bounds) :
    GrainXWindow(system, create_x_window(system.display(), system.screen(), 
					 parent, bounds))
{}

Window GrainXSubwindow::create_x_window(Display* display, int screen, 
					Window parent, const GrainRect& bounds)
{
    XSetWindowAttributes attributes;
    attributes.background_pixel=WhitePixel(display, screen);
    attributes.border_pixel=BlackPixel(display, screen);

    Window window_id=XCreateWindow(
        display, 
	/* parent window */ parent, 
	bounds.position.x, bounds.position.y, bounds.size.x, bounds.size.y,
        /* border width */ 0, 
	/* depth */ CopyFromParent,
	/* class */ CopyFromParent, 
	/* visual */ CopyFromParent,
	/* value mask */ CWBackPixel|CWBorderPixel,
	&attributes);

    return window_id;
}
